---Biblioteca para realizao de conexes TCP. 
--Utiliza co-rotinas de lua para simular multi-thread.
--Fonte: <a href="http://www.telemidia.puc-rio.br/~francisco/nclua/index.html">Tutorial de NCLua</a>

-- TODO:
-- * nao aceita `tcp.execute` reentrante

--Declara localmente mdulos e fune globais pois, ao definir
--o script como um mdulo, o acesso ao ambiente global  perdido
local _G, coroutine, event, assert, pairs, type, print
    = _G, coroutine, event, assert, pairs, type, print
local s_sub = string.sub

module 'tcp'

---Lista de conexes TCP ativas
local CONNECTIONS = {}

---Obtem a co-rotina em execuo
--@returns Retorna o identificador da co-rotina em execuo
local current = function ()
    return assert(CONNECTIONS[assert(coroutine.running())])
end

---(Re)Inicia a execuo de uma co-rotinas. Estas, so criadas
--suspensas, assim,  necessrio resum-las para entrarem
--em execuo.
--@param co Co-rotina a ser resumida
--@param ... Todos os parmetros adicionais
--so passados  funo que a co-rotina executa.
--Quando a co-rotina  suspensa com yield, ao ser resumida
--novamente, estes parmetros extras passados na chamada de resume
--so retornados pela yield. Isto  usado, por exemplo, na co-rotina da
--funo receive, para receber a resposta de uma requisio TCP. Assim,
--ao iniciar, co-rotina da funo  suspensa para que fique aguardando
--a resposta da requisio TCP. Quando a funo tratadora de eventos (handler)
--recebe os dados, ela resume a co-rotina da funo receive. Os dados
--recebidos so passados  funo resume, e estes so retornados pela funo
--yield depois que a co-rotina  reiniciada.
local resume = function (co, ...)
    assert(coroutine.status(co) == 'suspended')
    assert(coroutine.resume(co, ...))
    if coroutine.status(co) == 'dead' then
       CONNECTIONS[co] = nil
    end
end

---Funo tratadora de eventos. Utilizada para tratar 
--os eventos gerados pelas chamadas s funes da classe tcp.
--@param evt Tabela contendo os dados do evento capturado
function handler (evt)
    if evt.class ~= 'tcp' then return end

    if evt.type == 'connect' then
        for co, t in pairs(CONNECTIONS) do
            if (t.waiting == 'connect') and
               (t.host == evt.host) and (t.port == evt.port) then
                t.connection = evt.connection
                t.waiting = nil
                --Continua a execuo da co-rotina,
                --fazendo com que a funo connect, que causou
                --o disparo do evento connect, capturado
                --por esta funo (handler), seja finalizada.
                resume(co) 
                break
            end
        end
        return
    end

    if evt.type == 'disconnect' then
        for co, t in pairs(CONNECTIONS) do
            if t.waiting and
               (t.connection == evt.connection) then
                t.waiting = nil
                resume(co, nil, 'disconnected')
            end
        end
        return
    end

	--Evento disparado quando existem dados a serem recebidos
	--aps a chamada da funo receive.
    if evt.type == 'data' then
        for co, t in pairs(CONNECTIONS) do
            if (t.waiting == 'data') and
            (t.connection == evt.connection) then
                --O atributo value da tabela evt contm os dados
                --recebidos. Assim, continua a execuo da funo que disparou
                --este evento (funo receive). O valor de evt.value
                -- retornado pela funo coroutine.yield, chamada
                --dentro da funo receive (que ficou suspensa
                --aguardando os dados serem recebidos).
                --Desta forma, dentro da funo receive, o retorno
                --de coroutine.yield contm os dados recebidos.
                resume(co, evt.value)
            end
        end
        return
    end
end
event.register(handler)

---Funo que deve ser chamada para iniciar uma conexo TCP.
--@param f Funo que dever executar as rotinas
--para realizao de uma conexo TCP, envio de requisies
--e obteno de resposta.   
--@param ... Todos os parmetros adicionais
--so passados  funo que a co-rotina executa.
--@see resume
function execute (f, ...)
    resume(coroutine.create(f), ...)
    print("finalizou tcp.execute")
end

---Conecta em um servidor por meio do protocolo TCP.
--A funo s retorna quando a conexo for estabelecida.
--@param host Nome do host para conectar
--@param port Porta a ser usada para a conexo
function connect (host, port)
    local t = {
        host    = host,
        port    = port,
        waiting = 'connect'
    }
    CONNECTIONS[coroutine.running()] = t

    event.post {
        class = 'tcp',
        type  = 'connect',
        host  = host,
        port  = port,
    }
    
    --Suspende a execuo da co-rotina.
    --A funo atual (connect) s retorna quando
    --a co-rotina for resumida, o que ocorre
    --quando o evento connect  capturado
    --pela funo handler. 
    return coroutine.yield() 
end

---Fecha a conexo TCP e retorna imediatamente
function disconnect ()
    local t = current()
    event.post {
        class      = 'tcp',
        type       = 'disconnect',
        connection = assert(t.connection),
    }
end

---Envia uma requisio TCP ao servidor no qual se est conectado, e retorna imediatamente.
--@param value Mensagem a ser enviada ao servidor.
function send (value)
    local t = current()
    event.post {
        class      = 'tcp',
        type       = 'data',
        connection = assert(t.connection),
        value      = value,
    }
end


---Recebe resposta de uma requisio enviada previamente
--ao servidor.
--@param pattern Padro para recebimento dos dados.
--Se passado *a, todos os dados da resposta so 
--retornados de uma s vez, sem precisar fazer
--chamadas sucessivas a esta funo.
--Se omitido, os dados vo sendo retornados parcialmente,
--sendo necessrias vrias chamadas  funo.  
function receive (pattern)
    pattern = pattern or '' -- TODO: '*l'/number
    local t = current()
    t.waiting = 'data'
    t.pattern = pattern
    
    if s_sub(pattern, 1, 2) ~= '*a' then
        --Suspende a execuo da funo, at que
        --um bloco de dados seja recebido.
        --Ela s  resumida depois que 
        --a funo handler (tratadora de eventos)
        --receber um bloco de dados. Nesse momento,
        --a funo receive retorna o bloco de dados.   
        --Tendo entrado neste if, o parmetro pattern ser
        --diferente de '*a', logo, sero necessrias
        --vrias chamadas sucessvas a receive para obter
        --toda a resposta da requisio enviada previamente
        --por meio da funo send.
        --A funo receive retorna nil quando no houver
        --mais nada para ser retornado.
        return coroutine.yield()
    end
    
    --Chegando aqui,  porque o parmetro pattern  igual
    --a '*a', indicando que a funo s deve retornar depois
    --que toda a resposta da requisio enviada previamente,
    --por meio da funo send, tiver sido retornada.
    local all = ''
    while true do
        --Suspende a execuo da funo, at que
        --um bloco de dados seja recebido.
        --Ela s  resumida depois que 
        --a funo handler (tratadora de eventos)
        --receber um bloco de dados. Nesse momento,
        --a funo receive retorna o bloco de dados.   
        --Se o resultado for nil, a funo finaliza 
        --devolvendo todos os blocos de resposta recebidos,
        --concatenados. No sendo nil, a funo suspende a execuo
        --at receber novo bloco.
        local ret = coroutine.yield()
        if ret then
            all = all .. ret
        else
            return all
        end
    end
end
